/*
 Copyright 2015 Nicolas Melot

 This file is part of Crown.

 Crown is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Crown is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Crown. If not, see <http://www.gnu.org/licenses/>.

*/


#include <pelib/Algebra.hpp>

#ifndef ANNEALING_H
#define ANNEALING_H

using namespace pelib;

namespace pelib
{
	namespace crown
	{
		/** Compute energy consumption of input schedule using a simple energy model **/
		float
		quality_simple(const Algebra &schedule);
		/** Compute energy consumption of input schedule using an energy model taking the energy consumed by idle cores into account **/
		float
		quality_idle(const Algebra &schedule);
		/** Compute energy consumption of input schedule using an energy model considering idle cores at shut down **/
		float
		quality_off(const Algebra &schedule);
		/** Compute the energy consumption of a input schedule using a generic energy model parametrized with scalars alpha, alpha-static, zeta and epsilon **/
		float
		energy_static_dynamic_busy_idle_active(const Algebra &rec);
		float
		quality_generic(const Algebra &schedule);

		/** Compute a new allocation within the distance in number of task and number of cores allocated
			@param task Proportion, between 0 and 1, of the number of tasks whose allocation can change. At least one task is selected for change.
			@param core Proportion, between 0 and 1, of the maximal number of core that can be added or removed to tasks transformed. Each task transformed is added or removed at least one core.
		**/
		Algebra
		neighbor_allocation(const Algebra &solution, float tasks, float cores);
		/** Compute a neighbor solution where distance influence both the number of tasks transformed and the number of cores a task can be added or removed. **/
		Algebra
		neighbor_allocation_full(const Algebra &solution, float distance);
		/** Compute a neighbor solution where distance only influence the number of tasks transformed. Their allocation can only change by one core. **/
		Algebra
		neighbor_allocation_tasks(const Algebra &solution, float distance);
		/** Compute a neighbor solution where distance only influcen the number of core a task can be added or removed. Only one task is transformed. **/
		Algebra
		neighbor_allocation_cores(const Algebra &solution, float distance);
		/** Compute a neighbor allocation by influencing the minimal efficiency of tasks begin transformed. **/
		Algebra
		neighbor_efficiency(const Algebra &solution, float distance);

		/** Run simulating annealing and produce a copy of the initial solution and add the allocation vector wi **/
		Algebra
		annealing(const Algebra &initial, const Algebra &initial_best, float (*quality)(const Algebra &solution), Algebra (*neighbor)(const Algebra &solution, float distance), float temperature, float cooling, float final, int max_transform, int max_new_state, float distance);

		/** Returns time in the number of instruction as counter by a time complexity analysis, to run simulated annealing with the input problem and parameters **/
		float
		annealing_complexity(float temperature, float cooling, float final, int max_transform, int max_new_state, float distance);
	}
}

#endif
