/*
 Copyright 2015 Nicolas Melot

 This file is part of Pelib.

 Pelib is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Pelib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Pelib. If not, see <http://www.gnu.org/licenses/>.
*/

#include <pelib/Algebra.hpp>
#include <pelib/Scalar.hpp>
#include <pelib/AmplSolver.hpp>
#include <pelib/Vector.hpp>
#include <pelib/Matrix.hpp>
#include <pelib/AmplInput.hpp>

#include <crown/CrownILPIntegrated.hpp>
#include <crown/CrownScheduler.hpp>
#include <crown/CrownConfigBinary.hpp>
#include <crown/CrownException.hpp>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace std;
using namespace pelib;
using namespace crown;

extern char _binary_crown_integrated_mod_start;
extern char _binary_crown_integrated_mod_end;
extern char _binary_crown_integrated_run_start;
extern char _binary_crown_integrated_run_end;

Algebra
CrownILPIntegrated::solve(const Algebra &tg, const Algebra &pt, const pelib::Algebra &param, std::map<const std::basic_string<char>, double> &statistics) const
{
	map<const string, const Algebra> data;
	data.insert(pair<const string, const Algebra>(string("taskgraph.dat"), tg));
	data.insert(pair<const string, const Algebra>(string("platform.dat"), pt));
	data.insert(pair<const string, const Algebra>(string("parameters.dat"), param));
        
	// Build model and run scripts input streams
	stringstream model(string(&_binary_crown_integrated_mod_start, &_binary_crown_integrated_mod_end));
	stringstream run(string(&_binary_crown_integrated_run_start, &_binary_crown_integrated_run_end));
	// Run the solver
	const Algebra *solution = AmplSolver(model, string("model.mod"), run, data, this->showOutput, this->showError).solve(statistics);
	Algebra alg = *solution;

	return alg;
}

Schedule
CrownILPIntegrated::schedule(const Taskgraph &tg, const Platform &pt, const Algebra &param, map<const string, double> &statistics) const
{
	// Build input collection
	//map<const string, const Algebra> data;
	Algebra taskgraph = tg.buildAlgebra(pt);
  
	if(taskgraph.find<Vector<int, string> >("name") == NULL)
	{
		CrownException("Cannot find tasks' vector name in Taskgraph conversion to Algebra record.");
	}
	Vector<int, string> name = *taskgraph.find<Vector<int, string> >("name");
	taskgraph.remove("name");

	Algebra p = this->addCrownConfig(tg, pt, param, statistics);
	float time_config = 0;
	const Scalar<float> *time_config_scalar = p.find<Scalar<float> >("time_crown_config");
	if(time_config_scalar != NULL)
	{
		time_config = time_config_scalar->getValue();
	}

	//AmplInput(AmplInput::intFloatHandlers()).dump(cout, pt.buildAlgebra());
	// Solve the problem
	Algebra res = solve(taskgraph, pt.buildAlgebra(), p, statistics);

	map<const string, double>::iterator ii = statistics.find("time");
	if(ii != statistics.end())
	{
		double time = ii->second;
		statistics.erase(ii);
		statistics.insert(pair<const string, double>("time_schedule", time));
		statistics.insert(pair<const string, double>("time_config", time_config));
		statistics.insert(pair<const string, double>("time_global", time + time_config));
	}

	// Insert back tasks' names
	res.insert(new Vector<int, string>(name));

	if(statistics.find("feasible")->second != 0)
	{
		res = crownToSchedule(res);
		res = Schedule::addStartTime(res, tg, pt);
		Schedule sched("Crown ILP Integrated", tg.getName(), res);

		return sched;
	}
	else
	{
		return Schedule("Crown ILP Integrated", tg.getName(), Schedule::table());
	}
}

Schedule
CrownILPIntegrated::schedule(const Taskgraph &tg, const Platform &pt, map<const string, double> &statistics) const
{
	return schedule(tg, pt, param, statistics);
}

CrownILPIntegrated::CrownILPIntegrated(const CrownConfig* config, bool showOutput, bool showError) : CrownScheduler(config, showOutput, showError)
{
	/* Do nothing else */
}

CrownILPIntegrated::CrownILPIntegrated(const Algebra &param, const CrownConfig* config, bool showOutput, bool showError) : CrownScheduler(param, config, showOutput, showError)
{
	/* Do nothing else */
}

CrownILPIntegrated::CrownILPIntegrated(const Taskgraph &tg, const Platform &pt, const Algebra &param, const CrownConfig* config, bool showOutput, bool showError) : CrownScheduler(tg, pt, param, config, showOutput, showError)
{
	/* Do nothing else */
}

CrownILPIntegrated::CrownILPIntegrated(const CrownILPIntegrated &src) : CrownScheduler(src.tg, src.pt, src.param, src.config, src.showOutput, src.showError)
{
	/* Do nothing else */
}

string
CrownILPIntegrated::getShortDescription() const
{
	float alpha = this->param.find<Scalar<float> >("alpha")->getValue();
	float kappa = this->param.find<Scalar<float> >("kappa")->getValue();
	float eta = this->param.find<Scalar<float> >("eta")->getValue();
	float zeta = this->param.find<Scalar<float> >("zeta")->getValue();
	stringstream ss;

	const CrownConfig *config = this->config;
	if(config == NULL)
	{
		config = getDefaultConfig();
	}

	std::streamsize p = ss.precision();
	ss.precision(4);
	ss << "Cr.ILP.Int-" << config->getShortDescription() << "(a=" << alpha << ",k=" << kappa << ",e=" << eta << ",z=" << zeta << ")";
	ss.precision(p);

	if(this->config == NULL)
	{
		delete config;
	}
	return ss.str();
}

float
CrownILPIntegrated::complexity(const Algebra &problem) const
{
	return 0;
}

float
CrownILPIntegrated::complexity(const Taskgraph &tg, const Platform &pt, const Algebra &param) const
{
	return 0;
}

CrownILPIntegrated*
CrownILPIntegrated::clone() const
{
	return new CrownILPIntegrated(*this);
}

