/*
  Copyright 2015 Nicolas Melot

  This file is part of Pelib.

  Pelib is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Pelib is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Pelib. If not, see <http://www.gnu.org/licenses/>.
*/

#include <pelib/Algebra.hpp>
#include <pelib/Scalar.hpp>
#include <pelib/AmplSolver.hpp>
#include <pelib/Vector.hpp>
#include <pelib/Matrix.hpp>
#include <pelib/AmplInput.hpp>
#include <pelib/XMLSchedule.hpp>

#include <crown/CrownModular.hpp>
#include <crown/CrownAllocationFastest.hpp>
#include <crown/CrownMappingLTLG.hpp>
#include <crown/CrownScalingHeight.hpp>
#include <crown/CrownConfigBinary.hpp>
#include <crown/CrownException.hpp>

#include <crown/mapping.h>
#include <crown/scaling.h>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace std;
using namespace pelib;
using namespace crown;

Algebra
CrownModular::solve(const Algebra &tg, const Algebra &pt, const pelib::Algebra &param, std::map<const std::basic_string<char>, double> &stats) const
{
	Algebra input = tg;
	input = input.merge(pt);
	input = input.merge(param);
	input = this->alloc->allocate(input, stats);
	input = this->mapping->map(input, stats);
	input = this->scaling->scale(input, stats);
	return input;
}

Schedule
CrownModular::schedule(const Taskgraph &tg, const Platform &pt, const Algebra &param, map<const string, double> &statistics) const
{
	// Take task names apart
	Algebra taskgraph = tg.buildAlgebra(pt);
	Vector<int, string> name = *taskgraph.find<Vector<int, string> >("name");
	taskgraph.remove("name");

	Algebra p = param.merge(config->configure(tg, pt, param, statistics));
	p = solve(taskgraph, pt.buildAlgebra(), p, statistics);

	float time = 0;
	map<const string, double>::iterator i = statistics.find("time_allocation");
	if(i != statistics.end())
	{
		time += i->second;
	}
	i = statistics.find("time_mapping");
	if(i != statistics.end())
	{
		time += i->second;
	}
	i = statistics.find("time_scaling");
	if(i != statistics.end())
	{
		time += i->second;
	}
	i = statistics.find("time_global");
	if(i != statistics.end())
	{
		statistics.erase(i);
	}
	statistics.insert(pair<const string, double>("time_global", time));

	// Put back task names
	p.insert(new Vector<int, string>(name));

	float m = p.find<Scalar<float> >("m")->getValue();
	float M = p.find<Scalar<float> >("M")->getValue();
	if(m >= 0 || M <= 0)
	{
		statistics.insert(pair<const string, double>("feasible", 1));
		p = crownToSchedule(p);
		p = Schedule::addStartTime(p, tg, pt);
		Schedule sched(getShortDescription(), tg.getName(), p);
		// for(map<string, double>::iterator it = statistics.begin() ; it != statistics.end() ; ++it)
		// {
		//	 debug(it->first);
		//	 debug(it->second);
		// }
		return sched;
	}
	else
	{
		statistics.insert(pair<const string, double>("feasible", 0));
		return Schedule(this->getShortDescription(), tg.getName(), Schedule::table());
	}
}

Schedule
CrownModular::schedule(const Taskgraph &tg, const Platform &pt, map<const string, double> &statistics) const
{
	return schedule(tg, pt, param, statistics);
}

float
CrownModular::complexity(const Algebra &problem) const
{
	return CrownScheduler::complexity(tg, pt, param) + this->alloc->complexity(problem) + this->mapping->complexity(problem) + this->scaling->complexity(problem);
}

float
CrownModular::complexity(const Taskgraph &tg, const Platform &pt, const Algebra &param) const
{
	return CrownScheduler::complexity(tg,pt, param) + this->alloc->complexity(tg, pt, param, this->config) + this->mapping->complexity(tg, pt, param, this->config) + this->scaling->complexity(tg, pt, param, this->config);
}

CrownModular*
CrownModular::clone() const
{
	return new CrownModular(*this);
}

void
CrownModular::initialize(const CrownAllocation *alloc, const CrownMapping *mapping, const CrownScaling *scaling)
{
	if(alloc == NULL)
	{
		this->alloc = new CrownAllocationFastest(0, showOutput, showError);
	}
	else
	{
		this->alloc = alloc->clone();
	}

	if(mapping == NULL)
	{
		this->mapping = new CrownMappingLTLG(config, alloc, showOutput, showError);
	}
	else
	{
		this->mapping = mapping->clone();
	}

	if(scaling == NULL)
	{
		this->scaling = new CrownScalingHeight(config, alloc, mapping, showOutput, showError);
	}
	else
	{
		this->scaling = scaling->clone();
	}
}

CrownModular::CrownModular(const CrownConfig *config, const CrownAllocation *alloc, const CrownMapping *mapping, const CrownScaling *scaling, bool showOutput, bool showError) : CrownScheduler(config, showOutput, showError)
{
	initialize(alloc, mapping, scaling);

	// TODO: restore arbitrary configuration
	//delete this->config;
	//this->config = new CrownConfigBinary();
}

CrownModular::CrownModular(const Algebra &param, const CrownConfig *config, const CrownAllocation *alloc, const CrownMapping *mapping, const CrownScaling *scaling, bool showOutput, bool showError) : CrownScheduler(config, showOutput, showError)
{
	initialize(alloc, mapping, scaling);

	// TODO: restore arbitrary configuration
	//delete this->config;
	//this->config = new CrownConfigBinary();
}

CrownModular::CrownModular(const Taskgraph &tg, const Platform &pt, const Algebra &param, const CrownConfig*, const CrownAllocation *alloc, const CrownMapping *mapping, const CrownScaling *scaling, bool showOutput, bool showError) : CrownScheduler(config, showOutput, showError)
{
	initialize(alloc, mapping, scaling);

	// TODO: restore arbitrary configuration
	//delete this->config;
	//this->config = new CrownConfigBinary();
}

CrownModular::CrownModular(const CrownModular &src) : CrownScheduler(src.tg, src.pt, src.param, src.config, src.showOutput, src.showError)
{
	initialize(src.alloc, src.mapping, src.scaling);

	// TODO: restore arbitrary configuration
	//delete this->config;
	//this->config = new CrownConfigBinary();
}

CrownModular::~CrownModular()
{
	delete this->alloc;
	delete this->mapping;
	delete this->scaling;
}

string
CrownModular::getShortDescription() const
{
	float alpha = this->param.find<Scalar<float> >("alpha")->getValue();
	float kappa = this->param.find<Scalar<float> >("kappa")->getValue();
	float eta = this->param.find<Scalar<float> >("eta")->getValue();
	float zeta = this->param.find<Scalar<float> >("zeta")->getValue();
	stringstream ss;

	const CrownConfig *config = this->config;
	if(config == NULL)
	{
		config = getDefaultConfig();
	}

	std::streamsize p = ss.precision();
	ss.precision(4);
	string alloc = this->alloc->getShortDescription();
	string mapping = this->mapping->getShortDescription();
	string scaling = this->scaling->getShortDescription();
	string conf = this->config->getShortDescription();
	ss << "Cr-" << alloc << "-" << mapping << "-" << scaling << "-" << conf << "(a=" << alpha << ",k=" << kappa << ",e=" << eta << ",z=" << zeta << ")";
	ss.precision(p);

	if(this->config == NULL)
	{
		delete config;
	}

	return ss.str();
}

const Algebra*
CrownModular::solve() const
{
	map<const string, double> statistics;
	Algebra *sol = new Algebra(schedule(tg, pt, param, statistics).buildAlgebra());
	return sol;
}

const Algebra*
CrownModular::solve(map<const string, double> &statistics) const
{
	Algebra *sol = new Algebra(schedule(tg, pt, param, statistics).buildAlgebra());
	return sol;
}

