/*
 Copyright 2015 Nicolas Melot

 This file is part of Pelib.

 Pelib is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Pelib is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Pelib. If not, see <http://www.gnu.org/licenses/>.
*/

#include <pelib/Vector.hpp>

#include <pelib/Scalar.hpp>
#include <pelib/Vector.hpp>
#include <pelib/Set.hpp>

#include <crown/CrownScaling.hpp>
#include <crown/CrownAllocationFastest.hpp>
#include <crown/CrownMappingLTLG.hpp>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace pelib;
using namespace crown;
using namespace std;

void
CrownScaling::initialize(const CrownAllocation *alloc, const CrownMapping *mapping)
{
	if(alloc == NULL)
	{
		this->alloc = new CrownAllocationFastest(0, showOutput, showError);
	}
	else
	{
		this->alloc = alloc->clone();
	}

	if(mapping == NULL)
	{
		this->mapping = new CrownMappingLTLG(config, alloc, showOutput, showError);
	}
	else
	{
		this->mapping = mapping->clone();
	}
}

CrownScaling::CrownScaling(const CrownConfig *config, const CrownAllocation *alloc, const CrownMapping *mapping, bool showOutput, bool showError): CrownScheduler(config, showOutput, showError)
{
	initialize(alloc, mapping);
}

CrownScaling::CrownScaling(const Algebra &param, const CrownConfig *config, const CrownAllocation *alloc, const CrownMapping *mapping, bool showOutput, bool showError): CrownScheduler(param, config, showOutput, showError)
{
	initialize(alloc, mapping);
}

CrownScaling::CrownScaling(const Taskgraph &tg, const Platform &pt, const Algebra &param, const CrownConfig *config, const CrownAllocation *alloc, const CrownMapping *mapping, bool showOutput, bool showError): CrownScheduler(tg, pt, param, config, showOutput, showError)
{
	initialize(alloc, mapping);
}

CrownScaling::CrownScaling(const CrownScaling &src) : CrownScheduler(src)
{
	initialize(src.alloc, src.mapping);
}

Schedule
CrownScaling::schedule(const Taskgraph &tg, const Platform &pt, std::map<const string, double> &statistics) const
{
	return schedule(tg, pt, param, statistics);
}

Schedule
CrownScaling::schedule(const Taskgraph &tg, const Platform &pt, const Algebra &param, std::map<const string, double> &statistics) const
{
	// Take task names apart
	Algebra taskgraph = tg.buildAlgebra(pt);
	Vector<int, string> name = *taskgraph.find<Vector<int, string> >("name");
	taskgraph.remove("name");

	Algebra p = param.merge(config->configure(tg, pt, param, statistics));
	p = solve(taskgraph, pt.buildAlgebra(), p, statistics);

	// Put back task names
	p.insert(new Vector<int, string>(name));

	p = crownToSchedule(p);
	p = Schedule::addStartTime(p, tg, pt);
	Schedule sched(getShortDescription(), tg.getName(), p);
	return sched;
}

Algebra
CrownScaling::solve(const Algebra &tg, const Algebra &pt, const pelib::Algebra &param, std::map<const std::basic_string<char>, double> &stats) const
{
	Algebra input = tg;
	input = input.merge(pt);
	input = input.merge(param);

	input = this->alloc->allocate(input, stats);
	input = this->mapping->map(input, stats);
	input = this->scale(input, stats);

	// Make up a frequency vector with minimal frequency
	std::map<int, float> freq;
	for(int i = 1; i <= tg.find<Scalar<float> >("n")->getValue(); i++)
	{
		freq.insert(pair<int, float>(i, *pt.find<Set<float> >("F")->getValues().begin()));
	}
	Vector<int, float> frequency("frequency", freq);
	input.insert(&frequency);

	return input;
}

CrownScaling::~CrownScaling()
{
	/* Do nothing */
}

