/*
 Copyright 2015 Nicolas Melot

 This file is part of Crown.

 Crown is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Crown is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Crown. If not, see <http://www.gnu.org/licenses/>.

*/


#include <fstream>

#include <pelib/argument_parsing.hpp>
#include <pelib/dl.h>
#include <pelib/Scalar.hpp>

#include <crown/CrownCompositeConsolidation.hpp>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace std;
using namespace pelib;
using namespace crown;

#ifdef __cplusplus
extern "C" {
#endif

struct args
{
	double alpha, eta, zeta, kappa;
	vector<pelib_argument_stream> composite;
	bool showError, showOutput;
	
};
typedef struct args args_t;

static args_t
parse(char **args)
{
	args_t out;
	out.showError = false;
	out.showOutput = false;
	out.alpha = 3;
	out.zeta = 0.649;
	out.eta = 0.018611;
	out.kappa = 52.639161335;

	for(; args[0] != NULL; args++)
	{
		if(strcmp(args[0], "--show-stdout") == 0)
		{
			out.showOutput = true;
			continue;
		}
		if(strcmp(args[0], "--show-stderr") == 0)
		{
			out.showError = true;
			continue;
		}
		if(strcmp(args[0], "--alpha") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.alpha;
			continue;
		}
		if(strcmp(args[0], "--eta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.eta;
			continue;
		}
		if(strcmp(args[0], "--zeta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.zeta;
			continue;
		}
		if(strcmp(args[0], "--kappa") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.kappa;
			continue;
		}
		if(strcmp(args[0], "--composite") == 0)
		{
			args++;
			pelib_argument_stream compo;
			pelib_argument_stream_init(&compo);
			args += pelib_argument_stream_parse(args, &compo) - 1;
			out.composite.push_back(compo);
			continue;
		}
	}

	return out;
}

const CrownScheduler*
crown_scheduler(size_t argc, char **argv, const CrownScheduler *composite)
{
	args_t args = parse(argv);
	Algebra param;
	param.insert(new Scalar<float>("alpha", args.alpha));
	param.insert(new Scalar<float>("eta", args.eta));
	param.insert(new Scalar<float>("zeta", args.zeta));
	param.insert(new Scalar<float>("kappa", args.kappa));

	CrownScheduler *scheduler = NULL;
	for(vector<pelib_argument_stream>::iterator i = args.composite.begin(); i != args.composite.end(); i++)
	{
		pelib_argument_stream compo = *i;
		void *libCompo = load_lib(compo.library);
		const CrownScheduler* (*sched)(size_t argc, char **argv, const CrownScheduler*) = (const CrownScheduler* (*)(size_t argc, char **argv, const CrownScheduler*))load_function(libCompo, "crown_scheduler");
		const CrownScheduler *new_scheduler = sched(compo.argc, compo.argv, scheduler);
		if(scheduler != NULL)
		{
			delete scheduler;
		}
		scheduler = new_scheduler->clone();
		void (*del)(const CrownScheduler*) = (void (*)(const CrownScheduler*))load_function(libCompo, "crown_delete_scheduler");
		del(new_scheduler);
		destroy_lib(libCompo);
	}

	CrownCompositeConsolidation *final = new CrownCompositeConsolidation(param, scheduler, args.showOutput, args.showError);

	if(scheduler != NULL)
	{
		delete scheduler;
	}

	return final;
}

#ifdef __cplusplus
}
#endif

