/*
 Copyright 2015 Nicolas Melot

 This file is part of Crown.

 Crown is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Crown is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Crown. If not, see <http://www.gnu.org/licenses/>.

*/


#include <fstream>

#include <pelib/argument_parsing.hpp>
#include <pelib/dl.h>
#include <pelib/Scalar.hpp>

#include <crown/crown-scheduler.hpp>
#include <crown/CrownILPIntegrated.hpp>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace std;
using namespace pelib;
using namespace crown;

#ifdef __cplusplus
extern "C" {
#endif

struct args
{
	double alpha, eta, zeta, kappa;
	pelib_argument_stream lib;
	bool showError, showOutput;
	
};
typedef struct args args_t;

static args_t
parse(char **args)
{
	args_t out;
	out.showError = false;
	out.showOutput = false;
	out.alpha = 3;
	out.zeta = 0.649;
	out.eta = 0.018611;
	out.kappa = 52.639161335;
	//out.configSrc = BINARY;
	pelib_argument_stream_init(&out.lib);

	for(; args[0] != NULL; args++)
	{
		if(strcmp(args[0], "--show-stdout") == 0)
		{
			out.showOutput = true;
			continue;
		}
		if(strcmp(args[0], "--show-stderr") == 0)
		{
			out.showError = true;
			continue;
		}
		if(strcmp(args[0], "--alpha") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.alpha;
			continue;
		}
		if(strcmp(args[0], "--eta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.eta;
			continue;
		}
		if(strcmp(args[0], "--zeta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.zeta;
			continue;
		}
		if(strcmp(args[0], "--kappa") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.kappa;
			continue;
		}
		if(strcmp(args[0], "--config") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.lib) - 1;
			continue;
		}
	}

	return out;
}

const CrownScheduler*
crown_scheduler(size_t argc, char **argv, const CrownScheduler *composite)
{
	args_t args = parse(argv);
	Algebra param;
	param.insert(new Scalar<float>("alpha", args.alpha));
	param.insert(new Scalar<float>("eta", args.eta));
	param.insert(new Scalar<float>("zeta", args.zeta));
	param.insert(new Scalar<float>("kappa", args.kappa));

	CrownScheduler *scheduler;

	// Convert solver output to general schedule
	if(args.lib.library == NULL)
	{
		scheduler = new CrownILPIntegrated(param, NULL, args.showOutput, args.showError);
	}
	else
	{
		char *library = args.lib.library;
		void *libConfig = load_lib(library);
		CrownConfig* (*config)(size_t argc, char **argv) = (CrownConfig* (*)(size_t argc, char **argv))load_function(libConfig, "crown_config");
		CrownConfig* conf = config(args.lib.argc, args.lib.argv);
		scheduler = new CrownILPIntegrated(param, conf, args.showOutput, args.showError);
		void (*del)(CrownConfig *) = (void (*)(CrownConfig*))load_function(libConfig, "crown_delete_config");
		del(conf);
		destroy_lib(libConfig);
	}

	return scheduler;
}

#ifdef __cplusplus
}
#endif

