/*
 Copyright 2015 Nicolas Melot

 This file is part of Crown.

 Crown is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Crown is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Crown. If not, see <http://www.gnu.org/licenses/>.

*/

#include <pelib/Scalar.hpp>
#include <pelib/argument_parsing.hpp>
#include <pelib/scheduler.h>
#include <pelib/dl.h>

#include <crown/crown-scheduler.hpp>
#include <crown/CrownAllocation.hpp>
#include <crown/CrownMapping.hpp>
#include <crown/CrownScaling.hpp>
#include <crown/CrownModular.hpp>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace std;
using namespace pelib;
using namespace crown;

#ifdef __cplusplus
extern "C" {
#endif

struct args
{
	double alpha, eta, zeta, kappa;
	pelib_argument_stream alloc, mapping, scaling, config;
	bool showError, showOutput;
	
};
typedef struct args args_t;

static args_t
parse(char **args)
{
	args_t out;
	out.showError = false;
	out.showOutput = false;
	out.alpha = 3;
	out.zeta = 0.649;
	out.eta = 0.018611;
	out.kappa = 52.639161335;
	pelib_argument_stream_init(&out.alloc);
	pelib_argument_stream_init(&out.mapping);
	pelib_argument_stream_init(&out.scaling);
	pelib_argument_stream_init(&out.config);

	for(; args[0] != NULL; args++)
	{
		if(strcmp(args[0], "--show-stdout") == 0)
		{
			out.showOutput = true;
			continue;
		}
		if(strcmp(args[0], "--show-stderr") == 0)
		{
			out.showError = true;
			continue;
		}
		if(strcmp(args[0], "--alpha") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.alpha;
			continue;
		}
		if(strcmp(args[0], "--eta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.eta;
			continue;
		}
		if(strcmp(args[0], "--zeta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.zeta;
			continue;
		}
		if(strcmp(args[0], "--kappa") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.kappa;
			continue;
		}
		if(strcmp(args[0], "--config") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.config) - 1;
			continue;
		}
		if(strcmp(args[0], "--allocation") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.alloc) - 1;
			continue;
		}
		if(strcmp(args[0], "--mapping") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.mapping) - 1;
			continue;
		}
		if(strcmp(args[0], "--scaling") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.scaling) - 1;
			continue;
		}
	}

	return out;
}

const CrownScheduler*
crown_scheduler(size_t argc, char **argv, const CrownScheduler *composite)
{
	args_t args = parse(argv);

	void *libAlloc = load_lib(args.alloc.library);
	void *libMapping = load_lib(args.mapping.library);
	void *libScaling = load_lib(args.scaling.library);
	void *libConfig = load_lib(args.config.library);

	CrownAllocation* (*allocate)(size_t argc, char **argv);
	CrownMapping* (*map)(size_t argc, char **argv);
	CrownScaling* (*scal)(size_t argc, char **argv);
	CrownConfig* (*conf)(size_t argc, char **argv);

	Algebra param;
	param.insert(new Scalar<float>("alpha", args.alpha));
	param.insert(new Scalar<float>("eta", args.eta));
	param.insert(new Scalar<float>("zeta", args.zeta));
	param.insert(new Scalar<float>("kappa", args.kappa));

	CrownAllocation* alloc = NULL;
	CrownMapping* mapping = NULL;
	CrownScaling* scaling = NULL;
	CrownConfig *config = NULL;

	// Load crown allocation
	if(args.alloc.library != NULL)
	{
		void *libConfig = load_lib(args.alloc.library);
		allocate = (CrownAllocation* (*)(size_t argc, char **argv))load_function(libConfig, "crown_allocation");
		alloc = allocate(args.alloc.argc, args.alloc.argv);
	}

	// Load crown mapping
	if(args.mapping.library != NULL)
	{
		void *libConfig = load_lib(args.mapping.library);
		map = (CrownMapping* (*)(size_t argc, char **argv))load_function(libConfig, "crown_mapping");
		mapping = map(args.mapping.argc, args.mapping.argv);
	}

	// Load crown scaling
	if(args.scaling.library != NULL)
	{
		void *libConfig = load_lib(args.scaling.library);
		scal = (CrownScaling* (*)(size_t argc, char **argv))load_function(libConfig, "crown_scaling");
		scaling = scal(args.scaling.argc, args.scaling.argv);
	}

	// Load crown config
	if(args.config.library != NULL)
	{
		void *libConfig = load_lib(args.config.library);
		conf = (CrownConfig* (*)(size_t argc, char **argv))load_function(libConfig, "crown_config");
		config = conf(args.config.argc, args.config.argv);
	}

	CrownScheduler *scheduler;
	// Compute a schedule
	scheduler = new CrownModular(param, config, alloc, mapping, scaling, args.showOutput, args.showError);

	// Destroy object instances by libraries
	if(args.alloc.library != NULL)
	{
		void (*del)(CrownAllocation *) = (void (*)(CrownAllocation*))load_function(libAlloc, "crown_delete_algebra");
		del(alloc);
		destroy_lib(libAlloc);
	}

	// Destroy object instances by libraries
	if(args.mapping.library != NULL)
	{
		void (*del)(CrownMapping *) = (void (*)(CrownMapping*))load_function(libMapping, "crown_delete_algebra");
		del(mapping);
		destroy_lib(libMapping);
	}

	// Destroy object instances by libraries
	if(args.scaling.library != NULL)
	{
		void (*del)(CrownScaling *) = (void (*)(CrownScaling*))load_function(libScaling, "crown_delete_algebra");
		del(scaling);
		destroy_lib(libScaling);
	}

	// Destroy object instances by libraries
	if(args.config.library != NULL)
	{
		void (*del)(CrownConfig *) = (void (*)(CrownConfig*))load_function(libConfig, "crown_delete_config");
		del(config);
		destroy_lib(libConfig);
	}

	return scheduler;
}

#ifdef __cplusplus
}
#endif
