option show_stats;
option solver gurobi_ampl;

set F;
param Fin;
set Fi{i in 1..Fin};
param Funit;
param zeta;
param alpha;
param M;
param eta;
param kappa;

## Model parameters
param p; # Number of processors
param b = 2; # Child per node
param L = log(p) / log(b);

set P = 1..p; # Set of processors
set B = 1..b; # Set of child at each node except leaves

## Convenient variables
param q = 2 * p -1;
param crown_configuration_groups{i in 1..q, j in 1..p};
param crown_configuration_tree{i in 1..q, j in 1..q};

## Hierarchical structure of processor groups
set Pi{i in 1..q} within P = if floor(log(i) / log(b)) + 1 <= L then union{j in B} Pi[b * i + j - 1] else {i - b^L + 1};
set G{i in P} = union{j in 1..q} if i in Pi[j] then {j} else {};
param pi{i in 1..q} = card(Pi[i]);
## Matrix equivalent to pi (M + pi = Mpi)
param Mpi{i in 1..q, j in P} = if pi[i] = j then 1 else 0;

## Tasks
param n; # Number of tasks
param c{t1 in 1..n, t2 in 1..n}; ## Communication channels
set T = 1..n; # Task set
param Wi{i in T}; # Task width
param Tau{i in T}; # Task work
param e{i in T, j in P}; # Task efficiency distribution
param e_max{i in T, j in P} = if Wi[i] >= j then e[i, j] else min(e[i, j], (e[i, j] * Wi[i]) / j); # Maximal task efficiency after a task's given efficiency and width

## Resource allocation
#param wi{i in T} = max(1, min(Wi[i], p)); # Always give maximal resources a task can handle
param wi{i in T}; # Get resource allocation from mapping.dat

# Matrix equivalent to wi (M + wi = Mwi)
param Mwi{i in T, j in P} binary = if wi[i] = j then 1 else 0;

var y{i in T, j in 1..q} binary;
subject to exact_one_group_per_task{i in T}: sum{j in 1..q} y[i, j] = 1; # Exactly one number of cores per task
subject to map_y_to_wi {i in T}: wi[i] = sum{j in 1..q} pi[j] * y[i, j]; # A task cannot receive more cores than it has been allocated to at allocation step ( =: receives exactly as many cores as allocated; <=: a task receives at least as many cores as allocated)
var _yy{i in T} = sum{j in 1..q} y[i,j] * j;
var group{i in T} = round(_yy[i]); # make sure output is integer digits

## Parallel work for tasks mapped to groups
param pTau{i in T} = Tau[i] / (sum{j in P} Mwi[i, j] * wi[i] * e_max[i, j]);

## Mapping
var max_group_load;
var group_load {i in 1..q} = sum{j in T} pTau[j] * y[j, i];
#subject to upper_bound {i in 1..q}: group_load[i] <= max_group_load; # Balance tasks and load within each groups separatly
subject to optimal_balance {i in P}: sum{ii in G[i]} group_load[ii] <= max_group_load; # Balance tasks and load within all processors

var core_makespan {i in P} = sum{ii in G[i]} group_load[ii];
var maxload = max{i in P} core_makespan[i];

var quality = maxload;

minimize total: max_group_load;
