/*
 Copyright 2015 Nicolas Melot

 This file is part of Crown.

 Crown is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Crown is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Crown. If not, see <http://www.gnu.org/licenses/>.

*/


#include <fstream>

#include <pelib/Scalar.hpp>
#include <pelib/argument_parsing.hpp>
#include <pelib/dl.h>

#include <crown/crown-scheduler.hpp>
#include <crown/CrownMappingLTLG.hpp>

#ifdef debug
#undef debug
#endif

#define debug(var) cout << "[" << __FILE__ << ":" << __FUNCTION__ << ":" << __LINE__ << "] " << #var << " = \"" << (var) << "\"" << endl;

using namespace std;
using namespace pelib;
using namespace crown;

#ifdef __cplusplus
extern "C" {
#endif

struct args
{
	double alpha, eta, zeta, kappa;
	bool showError, showOutput;
	pelib_argument_stream alloc, config;
	
};
typedef struct args args_t;

static args_t
parse(char **args)
{
	args_t out;
	out.showError = false;
	out.showOutput = false;
	out.alpha = 3;
	out.zeta = 0.649;
	out.eta = 0.018611;
	out.kappa = 52.639161335;
	pelib_argument_stream_init(&out.alloc);
	pelib_argument_stream_init(&out.config);

	for(; args[0] != NULL; args++)
	{
		if(strcmp(args[0], "--show-stdout") == 0)
		{
			out.showOutput = true;
			continue;
		}
		if(strcmp(args[0], "--show-stderr") == 0)
		{
			out.showError = true;
			continue;
		}
		if(strcmp(args[0], "--alpha") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.alpha;
			continue;
		}
		if(strcmp(args[0], "--eta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.eta;
			continue;
		}
		if(strcmp(args[0], "--zeta") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.zeta;
			continue;
		}
		if(strcmp(args[0], "--kappa") == 0)
		{
			args++;
			stringstream str(args[0]);
			str >> out.kappa;
			continue;
		}
		if(strcmp(args[0], "--allocation") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.alloc) - 1;
			continue;
		}
		if(strcmp(args[0], "--config") == 0)
		{
			args++;
			args += pelib_argument_stream_parse(args, &out.config) - 1;
			continue;
		}
	}

	return out;
}

const CrownMapping*
crown_mapping(size_t argc, char **argv)
{
	args_t args = parse(argv);
	void *libAlloc = load_lib(args.alloc.library);
	void *libConfig = load_lib(args.config.library);
	CrownAllocation* (*allocate)(size_t argc, char **argv);
	CrownConfig* (*conf)(size_t argc, char **argv);
	CrownAllocation* alloc = NULL;
	CrownConfig *config = NULL;

	// Load crown allocation
	if(args.alloc.library != NULL)
	{
		void *libConfig = load_lib(args.alloc.library);
		allocate = (CrownAllocation* (*)(size_t argc, char **argv))load_function(libConfig, "crown_allocation");
		alloc = allocate(args.alloc.argc, args.alloc.argv);
	}

	// Load crown config
	if(args.config.library != NULL)
	{
		void *libConfig = load_lib(args.config.library);
		conf = (CrownConfig* (*)(size_t argc, char **argv))load_function(libConfig, "crown_config");
		config = conf(args.config.argc, args.config.argv);
	}

	// Convert solver output to general schedule
	CrownMapping *map = new CrownMappingLTLG(config, alloc, args.showOutput, args.showError);

	// Destroy object instances by libraries
	if(args.alloc.library != NULL)
	{
		void (*del)(CrownAllocation *) = (void (*)(CrownAllocation*))load_function(libAlloc, "crown_delete_algebra");
		del(alloc);
		destroy_lib(libAlloc);
	}

	// Destroy object instances by libraries
	if(args.config.library != NULL)
	{
		void (*del)(CrownConfig *) = (void (*)(CrownConfig*))load_function(libConfig, "crown_delete_config");
		del(config);
		destroy_lib(libConfig);
	}

	return map;
}

const CrownScheduler*
crown_scheduler(size_t argc, char **argv, const CrownScheduler *composite)
{
	return crown_mapping(argc, argv);
}

void
crown_delete_algebra(pelib::Algebra *alg)
{
	delete alg;
}

#ifdef __cplusplus
}
#endif

